# -*- coding: utf-8 -*-

from .toolbase import Tool
import os
import tempfile
import configparser
from pathlib import Path


class CppCheck(Tool):

    def default_config(self):
        cfg = configparser.ConfigParser()
        cfg[self.tool_name] = {
            "binary": self.tool_name,
            "file_blacklist": "",
            "jobs": 0,
            "default_includes": "",
            "supression_file": "",
            "default_args": "",
            "includes_as_system": "",
        }
        return cfg[self.tool_name]

    def suppression_file(self):
        if "suppression_file" in self.config and self.config["suppression_file"]:
            supp = Path(self.config["suppression_file"])
            if supp.exists():
                return str(supp.absolute())
        return None

    def includes(self, args):
        return filter(lambda arg: "-i" in arg, args[1:-1])

    def execute(self, cdb, args):
        result = 1
        if not self.tool_exists():
            raise EnvironmentError(f"tool: {self.tool_name} not in path, cannot execute.")

        all_includes = []
        all_sources = []
        for compilation_unit in cdb:
            directory = compilation_unit["directory"]
            temp_name_sources = None
            temp_name_includes = None
            filename = compilation_unit["file"]
            absolute_filename = (Path(directory) / filename).absolute()
            if os.path.isfile(absolute_filename) and self.should_scan(absolute_filename, args.file):
                all_includes.extend(self.includes(compilation_unit["command"]))
                all_sources.append(absolute_filename)

        with tempfile.NamedTemporaryFile(mode="w", delete=False) as t:
            try:
                temp_name_includes = t.name
                for line in list(set(all_includes)):
                    t.write(f"{line[:2]}\n")
                for line in self.default_includes():
                    t.write(f"{line}\n")
            except:
                pass

            total_files = 0
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as t:
                temp_name_sources = t.name
                for line in all_sources:
                    t.write(f"{line}\n")
                    total_files += 1

                arguments = [
                    f"--includes-file={temp_name_includes}",
                    f"--file-list={temp_name_sources}",
                    f"-j {self.max_tasks(args, total_files)}",
                ]
                arguments.extend(self.config.getlist("default_args"))
                suppressions = self.suppression_file()
                if suppressions:
                    arguments.append(f"--suppressions-list={suppressions}")

                if args.xml:
                    arguments.extend(["--xml", "--xml-version=2"])

                if os.path.isfile(temp_name_includes) and os.path.isfile(temp_name_sources):
                    tmp_cmd = f"{self.binary} {' '.join(arguments)}"
                    if args.output is not None:
                        final_command = f"{tmp_cmd} 2> {args.output}"
                    else:
                        final_command = tmp_cmd

                    result = self.run(final_command)

        return result



# -*- coding: utf-8 -*-

from xml.etree import ElementTree
from xml.etree.ElementTree import Element
from xml.dom import minidom
from xml.sax.saxutils import escape
from pathlib import Path
import tempfile
import os
import concurrent.futures
from functools import partial
import shutil
import signal
from .logger import LOGGER as log
from .toolbase import Tool
from .tidy_converter import OutputParser

list_of_futures = []


def inthandler(a1, a2):
    global list_of_futures
    for i in list_of_futures:
        i.cancel()


def process_cb(logline, future):
    log.debug(logline)


class ClangTidy(Tool):
    def convert_arguments(self, arg):
        if arg.startswith("/"):
            return arg.replace("/", "-", 1)
        return arg

    def format_output_to_xml(self, filename, allow_dupes=True):
        def prettify(elem):
            rough_string = ElementTree.tostring(elem, "utf-8")
            reparsed = minidom.parseString(rough_string)
            return reparsed.toprettyxml(indent=" " * 4)

        def generate_verbose(items):
            x = []
            for item in items if items else []:
                x.append(f"{item.path}:{item.line}:{item.column}: {item.message}")
            return x

        def clean_rules(rules):
            seen = []
            for rule in rules:
                if rule not in seen:
                    seen.append(rule)
            return seen

        p = OutputParser()

        xml_root = Element("results")
        xml_root.set("version", "2")
        xml_cppcheck = Element("cppcheck")
        xml_cppcheck.set("version", "1.77")
        xml_root.append(xml_cppcheck)
        xml_errors = Element("errors")
        rules = []
        rules = p.parse_messages_from_file(filename)
        if not allow_dupes:
            rules = clean_rules(rules)

        for rule in rules:
            xml_error = Element("error")
            xml_error.set("id", rule.checker)
            xml_error.set("severity", "warning")
            if "modernize-" in rule.checker or "google-readability-" in rule.checker or "readability-" in rule.checker:
                xml_error.set("severity", "style")
            elif "performance-" in rule.checker:
                xml_error.set("severity", "performance")
            message = escape(f"{rule.message} (l:{rule.line},c:{rule.column})")
            xml_error.set("msg", message)
            verbose = []
            verbose.extend(generate_verbose(rule.fixits))
            verbose.extend(generate_verbose(rule.notes))

            final = escape("\n".join(filter(None, verbose)))
            if len(final) > 0:
                xml_error.set("verbose", final)

            xml_location = Element("location")
            xml_location.set("file", rule.path)
            xml_location.set("line", str(rule.line))
            xml_error.append(xml_location)
            xml_errors.append(xml_error)

        xml_root.append(xml_errors)
        with open(filename, "w") as f:
            f.write(prettify(xml_root))

    def process_queue(self, cmd, tmp_dir):
        output = self.capture(cmd, True)
        if output[0] == "":
            output = output[1:]

        if output and tmp_dir:
            with tempfile.NamedTemporaryFile(mode="w", dir=tmp_dir, suffix=".log", delete=False) as t:
                t.write("\n".join(output))

    def _generate_cmd_queue(self, cdb, args):
        command_queue = []
        for compilation_unit in cdb:
            arguments = []

            directory = Path(compilation_unit["directory"]).absolute()
            full_command = compilation_unit["command"].split(" ")
            absolute_filename = directory / compilation_unit["file"]
            compiler = Path(full_command[0]).name.lower()

            defaults = self.config.getlist("default_args")

            extra = "--quiet"
            if args.vfsoverlay.is_file():
                extra = f"{extra} --vfsoverlay={args.vfsoverlay.resolve()}"

            if compiler == "cl.exe":
                extra = f"{extra} --extra-arg-before=--driver-mode=cl"

            if absolute_filename.is_file():
                if self.should_scan(absolute_filename, args.file):
                    arguments.extend(self.config.getlist("default_args"))
                    arguments.extend(self.quote_defines(full_command[1:]))
                    arguments = self.add_additions(arguments)
                    arguments = self.filter_arguments(arguments)
                    arguments = self.convert_includes(arguments)
                    arguments.extend(self.includes_as_cli_flags(self.default_includes()))
                    tmp_cmd = f"cd {directory} && {self.binary} {extra} {absolute_filename} -- {' '.join(arguments)}"
                    command_queue.append(tmp_cmd)
                else:
                    log.debug(f"File {absolute_filename} is not scanned")
        return command_queue

    def execute(self, cdb, args):
        result = 1
        if not self.tool_exists():
            raise EnvironmentError(f"tool: {self.tool_name} not in path, cannot execute.")

        command_queue = self._generate_cmd_queue(cdb, args)
        try:
            tmp_dir = None
            if args.output is not None:
                tmp_dir = Path(tempfile.mkdtemp())

            tasks = self.max_tasks(args, len(command_queue))
            global list_of_futures
            signal.signal(signal.SIGINT, inthandler)

            log.info("Starting scanning ...")
            with concurrent.futures.ProcessPoolExecutor(tasks) as executor:
                for cmd in command_queue:
                    future = executor.submit(self.process_queue, cmd, tmp_dir)
                    future.add_done_callback(partial(process_cb, cmd))
                    future = list_of_futures.append(future)

            log.info("Waiting for scanning processes to finish ..")
            concurrent.futures.wait(list_of_futures)
            log.info("Scanning done ...")

            if args.output is not None:
                log.debug(f"User requested output to {args.output}, scanning log files and saving ...")
                with open(args.output, "w") as dst:
                    for name in tmp_dir.glob("*.log"):
                        dst.write(name.read_text())

                if args.xml:
                    log.debug("User requested output as xml, converting ...")
                    self.format_output_to_xml(args.output, args.allow_dupes)
                shutil.rmtree(tmp_dir)
            log.info("All done")
            result = 0
        except KeyboardInterrupt:
            log.debug("Keyboard interrupt!")
            if tmp_dir is not None:
                shutil.rmtree(tmp_dir)
            os.kill(0, 9)
        except Exception as e:  # TODO: Add proper exception handling
            log.error(f"Exception: {e} .-- ")
            result = 1
        log.info("All done")
        return result



# -*- coding: utf-8 -*-

from .toolbase import Tool  # noqa: F401
from .clangtidy import ClangTidy  # noqa: F401
from .clang import Clang  # noqa: F401
from .lizard import Lizard  # noqa: F401
from .cppcheck import CppCheck  # noqa: F401

TOOLS = {
    "clang-tidy": ClangTidy,
    "lizard": Lizard,
    "cppcheck": CppCheck,
    # "clang": Clang,  DISABLE FOR NOW. Not working
}
__all__ = ["Tool", "ClangTidy", "Lizard", "CppCheck", "TOOLS"]



# -*- coding: utf-8 -*-

import sys
import json
import traceback
from .cdb_tools import TOOLS
from .misc import remove_untouched_files, remove_dupes, argument_parser
from configparser import ConfigParser
from .logger import LOGGER as log, LOG_LEVELS


def main():
    args = argument_parser(TOOLS).parse_args()
    log.setLevel(LOG_LEVELS[args.loglevel])
    cdb = None
    ret = 0
    processcdb_config = ConfigParser()

    if args.dumpconfigs:
        for tool_name in TOOLS:
            tool = TOOLS[tool_name](tool_name)
            processcdb_config[tool_name] = tool.default_config()
        config_file = args.config.absolute()
        config_file.parent.mkdir(parents=True, exist_ok=True)
        if config_file.exists():
            log.warn(f"Configuration file {config_file} already exists, overwriting")
        with config_file.open("w") as output:
            processcdb_config.write(output)
        log.info(f"Configuration file written to {config_file}")
        sys.exit(0)

    processcdb_config.read(args.config)
    try:
        tool = TOOLS[args.tool](args.tool, processcdb_config)
    except KeyError:
        log.error(f"Unknown tool '{args.tool}' - cant initilize")
        return 127

    if args.cdb.is_file():
        cdb = json.loads(args.cdb.read_text())
        if cdb:
            if args.commit_a is not None:
                cdb = remove_untouched_files(cdb, (args.commit_a, args.commit_b))

            if not args.allow_dupes:
                cdb = remove_dupes(cdb)

            try:
                ret = tool.execute(cdb, args)
                log.info(f"Return value from tool process: {ret}")
            except EnvironmentError as e:
                log.error(f"Cant process: {e}")
            except Exception as e:
                log.error(f"{e}")
                traceback.print_exc()
        else:
            log.error(f"File '{args.cdb}' is empty")
    else:
        log.error(f"File '{args.cdb}' does not exist")
    return ret  # TODO: Add proper return codes


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

import platform
import subprocess
from pathlib import Path
from whatthepatch import parse_patch
from git import Repo
from appdirs import AppDirs
from collections import ChainMap
import argparse
from ._version import get_versions
from .logger import LOGGER as log, LOG_LEVELS  # noqa: F401

__author__ = "Jani Mikkonen"
__email__ = "jani.mikkonen@gmail.com"
__version__ = get_versions()["version"]


def is_windows():
    return "WINDOWS" in platform.system().upper()


def is_cygwin():
    return "CYGWIN" in platform.system().upper()


def is_any_windows():
    return is_windows() or is_cygwin()


def is_mac():
    return "Darwin" in platform.system()


def capture_output(args, captureErr=True, capture_exceptions=True):
    buff = ""
    params = {"shell": True, "universal_newlines": True}
    if captureErr:
        params["stderr"] = subprocess.STDOUT
    try:
        buff = subprocess.check_output(args, **params)
    except subprocess.CalledProcessError as e:
        buff = e.stdout
        if capture_exceptions:
            with open("error.log", "a") as f:
                f.write(f"ARGUMENTS: {args}\n")
                f.write(f"STDOUT: {e.stdout}\n")
                f.write("---------------------------------------\n")
    return buff.split("\n")


def remove_untouched_files(cdb, commits):
    def modified_lines(data):
        return data[0] is None and data[1] is not None

    def line_numbers(data):
        return data[1]

    patch = None
    repo = Repo(".", search_parent_directories=True)
    commits = list(filter(None, commits))
    if len(commits) == 2:
        patch = parse_patch(repo.git.diff(commits[0], commits[1]))
    else:
        patch = parse_patch(repo.git.diff(f"{commits[0]}^"))
    new_cdb = []
    base_dir = Path(repo.git_dir).parent
    lookup = {}
    for i in cdb:
        current = Path(i["directory"]) / i["file"]
        lookup[current] = i

    for diff in patch:
        fullname = base_dir / diff.header.new_path
        fullname = (base_dir / diff.header.new_path).absolute()
        if fullname in lookup:
            cdb_entry = lookup[fullname]
            cdb_entry["changes_rows"] = map(line_numbers, filter(modified_lines, diff.changes))
            new_cdb.append(cdb_entry)
    return new_cdb


def remove_dupes(cdb):
    seen = []
    new_cdb = []
    for entry in cdb:
        current = Path(entry["directory"]) / entry["file"]
        if current not in seen:
            new_cdb.append(entry)
            seen.append(current)
    return new_cdb


def argument_parser(tools):
    # TODO: Offload tool specific argument to the tool class itself if possible.
    app_dirs = AppDirs("processcdb", __author__)
    default_config_file = Path(app_dirs.user_config_dir) / "processcdb.ini"
    parser = argparse.ArgumentParser(description="Static analysis wrapper", epilog=f"Available tools: \n{','.join(tools.keys())}")
    parser.add_argument(
        "--cdb",
        "-c",
        action="store",
        dest="cdb",
        metavar="f",
        type=Path,
        default=Path("compile_commands.json"),
        help="Full name with path to compile_commands.json one wishes to process",
    )
    parser.add_argument(
        "--tool",
        "-t",
        action="store",
        dest="tool",
        default="clang-tidy",
        metavar="toolname",
        type=str,
        help="Tool to use to process cdb file",
    )
    parser.add_argument(
        "--config",
        action="store",
        dest="config",
        metavar="f",
        default=default_config_file,
        type=Path,
        help=f"Absolute path to a file where default configurations are loaded from. Default: {default_config_file}",
    )
    parser.add_argument(
        "--vfsoverlay",
        action="store",
        dest="vfsoverlay",
        metavar="f",
        type=Path,
        default=Path("vfsoverlayjson"),
        help="Full name with path to possible vfsoverlay file to pass to clang-tidy.",
    )
    xml = parser.add_mutually_exclusive_group()
    xml.add_argument(
        "--xml", dest="xml", action="store_true", help="If the tool allows, generate report in XML, defaults to plaintext"
    )
    xml.add_argument("--no-xml", dest="xml", action="store_false")
    parser.set_defaults(xml=False)

    allow_dupes = parser.add_mutually_exclusive_group()
    allow_dupes.add_argument(
        "--dupes",
        dest="allow_dupes",
        action="store_true",
        help="Allow duplicates in cdb to be scanned. Defaults to no-dupes",
    )
    allow_dupes.add_argument("--no-dupes", dest="allow_dupes", action="store_false")
    parser.set_defaults(allow_dupes=False)

    parser.set_defaults(xml=False)
    parser.add_argument(
        "--output",
        "-o",
        action="store",
        dest="output",
        default=None,
        metavar="f",
        help="Where results will be stored. Depending on the tool, this is either directory or a explicit file",
    )
    parser.add_argument(
        "--jobs",
        "-j",
        type=int,
        dest="jobs",
        action="store",
        default=0,
        metavar="N",
        help="If chosen tool allows, use this many parallel processes. 0 for automatic detection, otherwise, limit to N",
    )
    parser.add_argument(
        "--commit-a",
        type=str,
        dest="commit_a",
        action="store",
        default=None,
        metavar="N",
        help="Limit scanning to particular commit^",
    )
    parser.add_argument(
        "--commit-b",
        type=str,
        dest="commit_b",
        action="store",
        default=None,
        metavar="N",
        help="If specified, scan between commit-a and commit-b, otherwise commit-a is commit-a^",
    )
    parser.add_argument(
        "file",
        nargs="*",
        type=Path,
        help="If specified, scan only these files if they are found in cdb. If files specified here are not absolute paths, they will be relative the working directory",
    )
    parser.add_argument(
        "--dump-configs",
        dest="dumpconfigs",
        default=False,
        action="store_true",
        help="Generates default configuration",
    )
    parser.add_argument(
        "-l", "--loglevel", default="info", dest="loglevel", choices=list(LOG_LEVELS.keys())[1:], help="Log Level"
    )
    return parser


def to_list(value):
    try:
        return value.split(";")
    except:
        return []


def to_dict(value):
    def format(val):
        k,v = val.split("=", 2)
        return {k: v.split(",")}
    try:
        return dict(ChainMap(*map(format, to_list(value))))
    except:
        return {}



# -*- coding: utf-8 -*-

from .toolbase import Tool
import tempfile
import os


class Lizard(Tool):
    def execute(self, cdb, args):
        result = 1
        temp_name = None
        if not self.tool_exists():
            raise EnvironmentError(f"tool: {self.tool_name} not in path, cannot execute.")

        total_files = 0
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as t:
            temp_name = t.name
            for compilation_unit in cdb:
                directory = compilation_unit["directory"]
                filename = compilation_unit["file"]
                absolute_filename = os.path.abspath(os.path.join(directory, filename))
                if os.path.isfile(absolute_filename) and self.should_scan(absolute_filename, args.file):
                    t.write(f"{absolute_filename}\n")
                    total_files += 1

            if os.path.isfile(temp_name):
                arguments = [
                    "-l cpp",
                    "--ignore_warnings -1",
                    f"--working_threads {self.max_tasks(args, total_files)}",
                    f"--input_file {temp_name} -ENS",
                ]
                if args.xml:
                    arguments.extend(["--xml"])

                tmp_cmd = f"lizard {' '.join(arguments)}"
                if args.output is not None:
                    final_command = f"{tmp_cmd} > {args.output}"
                else:
                    final_command = tmp_cmd
                result = self.run(final_command)

            return result



# -*- coding: utf-8 -*-

import os
import re


class Note(object):
    def __init__(self, path, line, column, message):
        self.path = path
        self.line = line
        self.column = column
        self.message = message

    def __eq__(self, other):
        return (
            self.path == other.path and self.line == other.line and self.column == other.column and self.message == other.message
        )

    def __str__(self):
        return "path=%s, line=%d, column=%s, message=%s" % (self.path, self.line, self.column, self.message)


class Message(Note):
    def __init__(self, path, line, column, message, checker, fixits=None, notes=None):
        super(Message, self).__init__(path, line, column, message)
        self.checker = checker
        self.fixits = fixits if fixits else []
        self.notes = notes if notes else []

    def __eq__(self, other):
        return (
            super(Message, self).__eq__(other) and self.checker == other.checker and self.fixits == other.fixits and self.notes == other.notes
        )

    def __str__(self):
        return "%s, checker=%s, fixits=%s, notes=%s" % (
            super(Message, self).__str__(),
            self.checker,
            [str(fixit) for fixit in self.fixits],
            [str(note) for note in self.notes],
        )


class OutputParser(object):
    # Regex for parsing a clang-tidy message.
    message_line_re = re.compile(
        # File path followed by a ':'.
        r"^(?P<path>[\S ]+?):"
        # Line number followed by a ':'.
        r"(?P<line>\d+?):"
        # Column number followed by a ':' and a space.
        r"(?P<column>\d+?): "
        # Severity followed by a ':'.
        r"(?P<severity>(error|warning)):"
        # Checker message.
        r"(?P<message>[\S \t]+)\s*"
        # Checker name.
        r"\[(?P<checker>.*)\]"
    )

    # Matches a note.
    note_line_re = re.compile(
        # File path followed by a ':'.
        r"^(?P<path>[\S ]+?):"
        # Line number followed by a ':'.
        r"(?P<line>\d+?):"
        # Column number followed by a ':' and a space.
        r"(?P<column>\d+?): "
        # Severity == note.
        r"note:"
        # Checker message.
        r"(?P<message>.*)"
    )

    def __init__(self):
        self.messages = []

    def parse_messages_from_file(self, path):
        with open(path, "r", encoding="utf-8", errors="ignore") as file:
            return self.parse_messages(file)

    def parse_messages(self, tidy_out):
        titer = iter(tidy_out)
        try:
            next_line = next(titer)
            while True:
                message, next_line = self._parse_message(titer, next_line)
                if message is not None:
                    self.messages.append(message)
        except StopIteration:
            pass

        return self.messages

    def _parse_message(self, titer, line):
        match = OutputParser.message_line_re.match(line)
        if match is None:
            return None, next(titer)

        message = Message(
            os.path.abspath(match.group("path")),
            int(match.group("line")),
            int(match.group("column")),
            match.group("message").strip(),
            match.group("checker").strip(),
        )

        try:
            line = next(titer)
            line = self._parse_code(message, titer, line)
            line = self._parse_fixits(message, titer, line)
            line = self._parse_notes(message, titer, line)

            return message, line
        except StopIteration:
            return message, ""

    @staticmethod
    def _parse_code(message, titer, line):
        # Eat code line.
        if OutputParser.note_line_re.match(line) or OutputParser.message_line_re.match(line):
            # LOG.debug("Unexpected line: %s. Expected a code line!", line)
            return line

        # Eat arrow line.
        # FIXME: range support?
        line = next(titer)
        if "^" not in line:
            # LOG.debug("Unexpected line: %s. Expected an arrow line!", line)
            return line
        return next(titer)

    @staticmethod
    def _parse_fixits(message, titer, line):
        while OutputParser.message_line_re.match(line) is None and OutputParser.note_line_re.match(line) is None:
            message_text = line.strip()

            if message_text != "":
                message.fixits.append(Note(message.path, message.line, line.find(message_text) + 1, message_text))
            line = next(titer)
        return line

    def _parse_notes(self, message, titer, line):
        while OutputParser.message_line_re.match(line) is None:
            match = OutputParser.note_line_re.match(line)
            if match is None:
                # LOG.debug("Unexpected line: %s", line)
                return next(titer)

            message.notes.append(
                Note(
                    os.path.abspath(match.group("path")),
                    int(match.group("line")),
                    int(match.group("column")),
                    match.group("message").strip(),
                )
            )
            line = next(titer)
            line = self._parse_code(message, titer, line)
        return line



# -*- coding: utf-8 -*-

from .misc import is_any_windows
from .toolbase import Tool


class Clang(Tool):
    def execute(self, cdb, args):
        if not self.tool_exists():
            raise EnvironmentError(f"{self.binary} not in path, cannot execute.")

        arguments = ["--force-analyze-debug-code", f"--cdb {args.cdb}"]
        if args.xml:
            arguments.extend(["--plist"])

        if args.output is not None:
            arguments.extend([f"--output {args.output}"])

        if is_any_windows():
            arguments.extend([f"--use-analyzer {self.binary}"])

        final_command = f"{self.binary} {' '.join(arguments)}"
        return self.run(final_command)



# -*- coding: utf-8 -*-

from .misc import is_any_windows, capture_output, to_dict, to_list
import multiprocessing
import subprocess
import configparser
import os
from .logger import LOGGER as log
import shutil
import shlex
from fnmatch import fnmatch


class Tool(object):
    tool_name = ""
    capture = staticmethod(capture_output)

    def default_config(self):
        converters = {"list": to_list, "dict": to_dict}
        cfg = configparser.ConfigParser(converters=converters)
        cfg[self.tool_name] = {
            "binary": self.tool_name,
            "file_blacklist": "",
            "arg_blacklist": "",
            "arg_additions": "",
            "jobs": 0,
            "default_includes": "",
            "default_args": "",
            "includes_as_system": "",
        }
        return cfg[self.tool_name]

    def _set_defaults(self, config):
        self.config = self.default_config()
        if config and self.tool_name in config:
            section = config[self.tool_name]
            for key in section.keys():
                self.config[key] = section[key]

    def __init__(self, tool_name, tool_config=None):
        self.tool_name = tool_name
        self.config = {}
        self._set_defaults(tool_config)

    @property
    def prevent_scan(self):
        if self.config and "file_blacklist" in self.config:
            return list(filter(None, self.config.getlist("file_blacklist")))
        return []

    @property
    def binary(self):
        tool_name = self.tool_name
        if "binary" in self.config and self.config["binary"]:
            tool_name = self.config["binary"]

        if is_any_windows():
            if " " in tool_name:
                tool_name = f'"{tool_name}"'  # Why oh why doesnt shlex.quote() use " on windows wtf?
        else:
            tool_name = shlex.quote(tool_name)
        return tool_name

    def max_tasks(self, args, total_tasks=0):
        default_jobs = self.config.getint("jobs")
        cpu_count = multiprocessing.cpu_count()
        if args.jobs != 0:
            default_jobs = args.jobs

        if default_jobs == 0:
            default_jobs = cpu_count

        if total_tasks > 0 and total_tasks < cpu_count:
            return total_tasks
        if default_jobs > cpu_count:
            log.warning(f"Requesting {default_jobs} concurrent processing jobs is higher than current {cpu_count} core count ")
        return default_jobs

    def default_includes(self):
        def clean(str):
            return str.replace("(framework directory)", "").strip()

        includes = []
        includes_from_config = self.config.getlist("default_includes")

        if includes_from_config[0] == "env":
            includes = os.environ.get("INCLUDE", "").split(";")
        elif includes_from_config[0] == "detect" and not is_any_windows():
            results = self.capture("clang -x c++ -v -E /dev/null")
            sidx = results.index("#include <...> search starts here:") + 1
            eidx = results.index("End of search list.")
            includes = map(lambda str: clean(str), results[sidx:eidx])

        return list(filter(None, includes))

    def tool_exists(self):
        return shutil.which(self.config["binary"]) is not None

    def quote_defines(self, args):
        # TODO: Maybe use shlex.quote here ?
        def fix(str):
            if str.startswith("-D"):
                return str.replace('"', '\\"')
            else:
                return str

        return map(lambda arg: fix(arg), args)

    def should_scan(self, filename, files):
        # Needs a rewrite
        res = False
        patterns = files if isinstance(files, list) else [files]
        blacklisted = any([fnmatch(filename, pattern) for pattern in self.prevent_scan])
        requested = any([fnmatch(filename, pattern) for pattern in patterns])
        if requested:
            if blacklisted:
                log.debug(f"File {filename} is blacklisted but requested to be scanned.")
            res = True
        else:
            if not patterns and not blacklisted:
                res = True
        return res

    def execute(self, cdb, args=None):
        return None

    def run(self, command_line):
        log.debug(f"RUN: {command_line}")
        return subprocess.call(command_line, shell=True)

    def include_string(self, include_path, path_matchers):
        if any([fnmatch(include_path, pattern) for pattern in path_matchers]):
            return f'-isystem "{include_path}"'
        else:
            return f"-I{include_path}"

    def includes_as_cli_flags(self, includes):
        path_matchers = self.config.getlist("includes_as_system")
        return list(map(lambda i: self.include_string(i, path_matchers), includes))

    def convert_includes(self, arguments):
        def convert(arg, path_matchers):
            if arg and arg.startswith("-I"):
                return self.include_string(arg[2:], path_matchers)
            return arg

        path_matchers = self.config.getlist("includes_as_system")
        return list(map(lambda arg: convert(arg, path_matchers), arguments))


    def filter_arguments(self, args):
        def allowed_argument(arg):
            return arg[1:] not in self.config.getlist("arg_blacklist")

        return list(filter(allowed_argument, args))

    def add_additions(self, args):
        new_args = args.copy()
        additions = self.config.getdict("arg_additions")
        for arg in args:
            key = arg[1:]
            if key in additions:
                new_args.extend(additions[key])
        return new_args



# -*- coding: utf-8 -*-

"""
processcdb
"""
import versioneer
from os.path import abspath, dirname, join
from setuptools import setup
PACKAGE_NAME = "processcdb"
CWD = abspath(dirname(__file__))
with open(join(CWD, "requirements.txt"), encoding="utf-8") as f:
    REQUIREMENTS = f.read().splitlines()

# Get the long description from the README file
with open(join(CWD, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

CLASSIFIERS = """
Development Status :: 3 - Alpha
Topic :: Software Development :: Testing
Operating System :: OS Independent
License :: OSI Approved :: Apache Software License
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Topic :: Software Development :: Testing
""".strip().splitlines()

setup(
    name=PACKAGE_NAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Front-end to process compile_commands.json file and run various static analysis tools against it",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rasjani/processcdb",
    author="Jani Mikkonen",
    author_email="jani.mikkonen@gmail.com",
    license="Apache License 2.0",
    classifiers=CLASSIFIERS,
    install_requires=REQUIREMENTS,
    keywords="staticanalysis frontend clang cppcheck clangtidy compile_commands.json",
    platforms="any",
    packages=[PACKAGE_NAME],
    package_dir={"": "src"},
    entry_points={
        'console_scripts': ['processcdb = processcdb.__main__:main'],
    }
)



# -*- coding: utf-8 -*-

import logging


LOG_LEVELS = {
    "notset": logging.NOTSET,
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
_APPNAME = "processcdb"
logging.basicConfig(format="[%(levelname)s]: %(message)s", level=logging.INFO)
LOGGER = logging.getLogger(_APPNAME)


def log_decorator(wrapped):
    """Decorator helper that logs function calls"""

    def log_enter_exit(*args, **kwargs):
        arguments = ""
        LOGGER.debug("{}({}) [ENTERING]".format(wrapped.__name__, arguments))
        result = wrapped(*args, **kwargs)
        LOGGER.debug("{}() [LEAVING]".format(wrapped.__name__))
        return result  # noqa: R504

    return log_enter_exit
